@@ -19,30 +19,39 @@
 //! # fn slow_code(){}
 //! ```
 
-// BEGIN - Embark standard lints v0.3
+// BEGIN - Embark standard lints v0.4
 // do not change or add/remove here, but one can add exceptions after this section
 // for more info see: <https://github.com/EmbarkStudios/rust-ecosystem/issues/59>
 #![deny(unsafe_code)]
 #![warn(
     clippy::all,
     clippy::await_holding_lock,
+    clippy::char_lit_as_u8,
+    clippy::checked_conversions,
     clippy::dbg_macro,
     clippy::debug_assert_with_mut_call,
     clippy::doc_markdown,
     clippy::empty_enum,
     clippy::enum_glob_use,
     clippy::exit,
+    clippy::expl_impl_clone_on_copy,
+    clippy::explicit_deref_methods,
     clippy::explicit_into_iter_loop,
+    clippy::fallible_impl_from,
     clippy::filter_map_next,
+    clippy::float_cmp_const,
     clippy::fn_params_excessive_bools,
     clippy::if_let_mutex,
+    clippy::implicit_clone,
     clippy::imprecise_flops,
     clippy::inefficient_to_string,
+    clippy::invalid_upcast_comparisons,
     clippy::large_types_passed_by_value,
     clippy::let_unit_value,
     clippy::linkedlist,
     clippy::lossy_float_literal,
     clippy::macro_use_imports,
+    clippy::manual_ok_or,
     clippy::map_err_ignore,
     clippy::map_flatten,
     clippy::map_unwrap_or,
@@ -51,25 +60,36 @@
     clippy::match_wildcard_for_single_variants,
     clippy::mem_forget,
     clippy::mismatched_target_os,
+    clippy::mut_mut,
+    clippy::mutex_integer,
     clippy::needless_borrow,
     clippy::needless_continue,
     clippy::option_option,
-    clippy::pub_enum_variant_names,
+    clippy::path_buf_push_overwrite,
+    clippy::ptr_as_ptr,
     clippy::ref_option_ref,
     clippy::rest_pat_in_fully_bound_structs,
+    clippy::same_functions_in_if_condition,
+    clippy::semicolon_if_nothing_returned,
     clippy::string_add_assign,
     clippy::string_add,
+    clippy::string_lit_as_bytes,
     clippy::string_to_string,
-    clippy::suboptimal_flops,
     clippy::todo,
+    clippy::trait_duplication_in_bounds,
     clippy::unimplemented,
     clippy::unnested_or_patterns,
     clippy::unused_self,
+    clippy::useless_transmute,
     clippy::verbose_file_reads,
+    clippy::zero_sized_map_values,
     future_incompatible,
     nonstandard_style,
     rust_2018_idioms
 )]
+// END - Embark standard lints v0.4
+// crate-specific exceptions:
+#![deny(missing_crate_level_docs)]
 
 mod data;
 mod merge;
@@ -102,7 +122,10 @@ pub type NanoSecond = i64;
 
 /// Stream of profiling events from one thread.
 #[derive(Clone, Default)]
-#[cfg_attr(feature = "with_serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(
+    feature = "serialization",
+    derive(serde::Deserialize, serde::Serialize)
+)]
 pub struct Stream(Vec<u8>);
 
 impl Stream {
@@ -169,7 +192,10 @@ pub struct Scope<'s> {
 
 /// Used to identify one source of profiling data.
 #[derive(Clone, Debug, Eq, PartialEq, Ord, PartialOrd)]
-#[cfg_attr(feature = "with_serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(
+    feature = "serialization",
+    derive(serde::Deserialize, serde::Serialize)
+)]
 pub struct ThreadInfo {
     /// Useful for ordering threads.
     pub start_time_ns: Option<NanoSecond>,
@@ -181,21 +207,53 @@ pub struct ThreadInfo {
 
 pub type FrameIndex = u64;
 
-/// A non-empty `Stream` plus some info about it.
+/// A `Stream` plus some info about it.
 #[derive(Clone)]
-#[cfg_attr(feature = "with_serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(
+    feature = "serialization",
+    derive(serde::Deserialize, serde::Serialize)
+)]
 pub struct StreamInfo {
+    /// The raw profile data.
     pub stream: Stream,
+
+    /// Total number of scopes in the stream.
     pub num_scopes: usize,
+
+    /// The depth of the deepest scope.
+    /// `0` mean no scopes, `1` some scopes without children, etc.
     pub depth: usize,
+
+    /// The smallest and largest nanosecond value in the stream.
+    ///
+    /// The default value is `(NanoSecond::MAX, NanoSecond::MIN)` which indicates an empty stream.
     pub range_ns: (NanoSecond, NanoSecond),
 }
 
+impl Default for StreamInfo {
+    fn default() -> Self {
+        Self {
+            stream: Default::default(),
+            num_scopes: 0,
+            depth: 0,
+            range_ns: (NanoSecond::MAX, NanoSecond::MIN),
+        }
+    }
+}
+
 impl StreamInfo {
-    fn new(stream: Stream) -> Result<StreamInfo> {
+    /// Parse a stream to count the depth, number of scopes in it etc.
+    ///
+    /// Try to avoid calling this, and instead keep score while collecting a `StreamInfo`.
+    pub fn parse(stream: Stream) -> Result<StreamInfo> {
         let top_scopes = Reader::from_start(&stream).read_top_scopes()?;
         if top_scopes.is_empty() {
-            Err(Error::Empty)
+            Ok(StreamInfo {
+                stream,
+                num_scopes: 0,
+                depth: 0,
+                range_ns: (NanoSecond::MAX, NanoSecond::MIN),
+            })
         } else {
             let (num_scopes, depth) = Reader::count_scope_and_depth(&stream)?;
             let min_ns = top_scopes.first().unwrap().record.start_ns;
@@ -209,13 +267,24 @@ impl StreamInfo {
             })
         }
     }
+
+    pub fn append(&mut self, other: Self) {
+        self.stream.append(other.stream);
+        self.num_scopes += other.num_scopes;
+        self.depth = self.depth.max(other.depth);
+        self.range_ns.0 = self.range_ns.0.min(other.range_ns.0);
+        self.range_ns.1 = self.range_ns.1.max(other.range_ns.1);
+    }
 }
 
 // ----------------------------------------------------------------------------
 
 /// One frame worth of profile data, collected from many sources.
 #[derive(Clone)]
-#[cfg_attr(feature = "with_serde", derive(serde::Deserialize, serde::Serialize))]
+#[cfg_attr(
+    feature = "serialization",
+    derive(serde::Deserialize, serde::Serialize)
+)]
 pub struct FrameData {
     pub frame_index: FrameIndex,
     pub thread_streams: BTreeMap<ThreadInfo, Arc<StreamInfo>>,
@@ -227,11 +296,11 @@ pub struct FrameData {
 impl FrameData {
     pub fn new(
         frame_index: FrameIndex,
-        thread_streams: BTreeMap<ThreadInfo, Stream>,
+        thread_streams: BTreeMap<ThreadInfo, StreamInfo>,
     ) -> Result<Self> {
         let thread_streams: BTreeMap<_, _> = thread_streams
             .into_iter()
-            .filter_map(|(info, stream)| Some((info, Arc::new(StreamInfo::new(stream).ok()?))))
+            .map(|(info, stream_info)| (info, Arc::new(stream_info)))
             .collect();
 
         let mut num_bytes = 0;
@@ -239,11 +308,11 @@ impl FrameData {
 
         let mut min_ns = NanoSecond::MAX;
         let mut max_ns = NanoSecond::MIN;
-        for stream in thread_streams.values() {
-            num_bytes += stream.stream.len();
-            num_scopes += stream.num_scopes;
-            min_ns = min_ns.min(stream.range_ns.0);
-            max_ns = max_ns.max(stream.range_ns.1);
+        for stream_info in thread_streams.values() {
+            num_bytes += stream_info.stream.len();
+            num_scopes += stream_info.num_scopes;
+            min_ns = min_ns.min(stream_info.range_ns.0);
+            max_ns = max_ns.max(stream_info.range_ns.1);
         }
 
         if min_ns <= max_ns {
@@ -263,21 +332,82 @@ impl FrameData {
         let (min, max) = self.range_ns;
         max - min
     }
+
+    /// Writes one Frame into a stream, prefixed by it's length (u32 le).
+    #[cfg(feature = "serialization")]
+    pub fn write_into(&self, write: &mut impl std::io::Write) -> anyhow::Result<()> {
+        use bincode::Options as _;
+        let serialized = bincode::options().serialize(self)?;
+        let compressed = lz4_flex::compress_prepend_size(&serialized);
+        write.write_all(b"PFD0")?;
+        write.write_all(&(compressed.len() as u32).to_le_bytes())?;
+        write.write_all(&compressed)?;
+        Ok(())
+    }
+
+    /// Read the next [`FrameData`] from a stream.
+    ///
+    /// `None` is returned if the end of the stream is reached (EOF),
+    /// or an end-of-stream sentinel of 0u32 is read.
+    #[cfg(feature = "serialization")]
+    pub fn read_next(read: &mut impl std::io::Read) -> anyhow::Result<Option<Self>> {
+        use anyhow::Context as _;
+
+        let mut header = [0_u8; 4];
+        if let Err(err) = read.read_exact(&mut header) {
+            if err.kind() == std::io::ErrorKind::UnexpectedEof {
+                return Ok(None);
+            } else {
+                return Err(err.into());
+            }
+        }
+
+        if header == [0_u8; 4] {
+            Ok(None) // end-of-stream sentinel.
+        } else if &header == b"PFD0" {
+            let mut compressed_length = [0_u8; 4];
+            read.read_exact(&mut compressed_length)?;
+            let mut compressed = vec![0_u8; u32::from_le_bytes(compressed_length) as usize];
+            read.read_exact(&mut compressed)?;
+
+            let serialized =
+                lz4_flex::decompress_size_prepended(&compressed).context("lz4 decompress")?;
+
+            use bincode::Options as _;
+            Ok(Some(
+                bincode::options()
+                    .deserialize(&serialized)
+                    .context("bincode deserialize")?,
+            ))
+        } else {
+            // Old packet without magic header
+            let mut bytes = vec![0_u8; u32::from_le_bytes(header) as usize];
+            read.read_exact(&mut bytes)?;
+
+            use bincode::Options as _;
+            Ok(Some(
+                bincode::options()
+                    .deserialize(&bytes)
+                    .context("bincode deserialize")?,
+            ))
+        }
+    }
 }
 
 // ----------------------------------------------------------------------------
 
 type NsSource = fn() -> NanoSecond;
-type ThreadReporter = fn(ThreadInfo, Stream);
+type ThreadReporter = fn(ThreadInfo, StreamInfo);
 
 /// Report a stream of profile data from a thread to the `GlobalProfiler` singleton.
-pub fn global_reporter(info: ThreadInfo, stream: Stream) {
-    GlobalProfiler::lock().report(info, stream)
+pub fn global_reporter(info: ThreadInfo, stream_info: StreamInfo) {
+    GlobalProfiler::lock().report(info, stream_info);
 }
 
 /// Collects profiling data for one thread
 pub struct ThreadProfiler {
-    stream: Stream,
+    stream_info: StreamInfo,
+    /// Current depth.
     depth: usize,
     now_ns: NsSource,
     reporter: ThreadReporter,
@@ -287,7 +417,7 @@ pub struct ThreadProfiler {
 impl Default for ThreadProfiler {
     fn default() -> Self {
         Self {
-            stream: Default::default(),
+            stream_info: Default::default(),
             depth: 0,
             now_ns: crate::now_ns,
             reporter: global_reporter,
@@ -318,16 +448,26 @@ impl ThreadProfiler {
         self.start_time_ns = Some(self.start_time_ns.unwrap_or(now_ns));
 
         self.depth += 1;
-        self.stream.begin_scope(now_ns, id, location, data)
+
+        self.stream_info.range_ns.0 = self.stream_info.range_ns.0.min(now_ns);
+        self.stream_info
+            .stream
+            .begin_scope(now_ns, id, location, data)
     }
 
     pub fn end_scope(&mut self, start_offset: usize) {
+        let now_ns = (self.now_ns)();
+        self.stream_info.depth = self.stream_info.depth.max(self.depth);
+        self.stream_info.num_scopes += 1;
+        self.stream_info.range_ns.1 = self.stream_info.range_ns.1.max(now_ns);
+
         if self.depth > 0 {
             self.depth -= 1;
         } else {
             eprintln!("puffin ERROR: Mismatched scope begin/end calls");
         }
-        self.stream.end_scope(start_offset, (self.now_ns)());
+
+        self.stream_info.stream.end_scope(start_offset, now_ns);
 
         if self.depth == 0 {
             // We have no open scopes.
@@ -336,8 +476,8 @@ impl ThreadProfiler {
                 start_time_ns: self.start_time_ns,
                 name: std::thread::current().name().unwrap_or_default().to_owned(),
             };
-            let stream = std::mem::take(&mut self.stream);
-            (self.reporter)(info, stream);
+            let stream_info = std::mem::take(&mut self.stream_info);
+            (self.reporter)(info, stream_info);
         }
     }
 
@@ -373,10 +513,16 @@ impl Ord for OrderedData {
     }
 }
 
+/// Add these to [`GlobalProfiler`] with [`GlobalProfiler::add_sink`].
+pub type FrameSink = Box<dyn Fn(Arc<FrameData>) + Send>;
+
+#[derive(Clone, Copy, Debug, Hash, PartialEq, Eq)]
+pub struct FrameSinkId(u64);
+
 /// Singleton. Collects profiling data from multiple threads.
 pub struct GlobalProfiler {
     current_frame_index: FrameIndex,
-    current_frame: BTreeMap<ThreadInfo, Stream>,
+    current_frame: BTreeMap<ThreadInfo, StreamInfo>,
 
     /// newest first
     recent_frames: std::collections::VecDeque<Arc<FrameData>>,
@@ -384,6 +530,9 @@ pub struct GlobalProfiler {
 
     slowest_frames: std::collections::BinaryHeap<OrderedData>,
     max_slow: usize,
+
+    next_sink_id: FrameSinkId,
+    sinks: std::collections::HashMap<FrameSinkId, FrameSink>,
 }
 
 impl Default for GlobalProfiler {
@@ -398,6 +547,8 @@ impl Default for GlobalProfiler {
             max_recent,
             slowest_frames: std::collections::BinaryHeap::with_capacity(max_slow),
             max_slow,
+            next_sink_id: FrameSinkId(1),
+            sinks: Default::default(),
         }
     }
 }
@@ -434,6 +585,10 @@ impl GlobalProfiler {
 
     /// Manually add frame data.
     pub fn add_frame(&mut self, new_frame: Arc<FrameData>) {
+        for sink in self.sinks.values() {
+            sink(new_frame.clone());
+        }
+
         let add_to_slowest = if self.slowest_frames.len() < self.max_slow {
             true
         } else if let Some(fastest_of_the_slow) = self.slowest_frames.peek() {
@@ -449,6 +604,15 @@ impl GlobalProfiler {
             }
         }
 
+        if let Some(last) = self.recent_frames.back() {
+            if new_frame.frame_index != last.frame_index + 1 {
+                // Keep `recent_frames` consecutive.
+                // Important when loading .puffin files which has both
+                // the slowest frames and most recent frames in the same stream.
+                self.recent_frames.clear();
+            }
+        }
+
         self.recent_frames.push_back(new_frame);
         while self.recent_frames.len() > self.max_recent {
             self.recent_frames.pop_front();
@@ -456,8 +620,11 @@ impl GlobalProfiler {
     }
 
     /// Report some profiling data. Called from `ThreadProfiler`.
-    pub fn report(&mut self, info: ThreadInfo, stream: Stream) {
-        self.current_frame.entry(info).or_default().append(stream);
+    pub fn report(&mut self, info: ThreadInfo, stream_info: StreamInfo) {
+        self.current_frame
+            .entry(info)
+            .or_default()
+            .append(stream_info);
     }
 
     /// The latest fully captured frame of data.
@@ -490,7 +657,7 @@ impl GlobalProfiler {
 
     /// How many frames of recent history to store.
     pub fn set_max_history(&mut self, max_recent: usize) {
-        self.max_recent = max_recent
+        self.max_recent = max_recent;
     }
 
     /// How many slow "spike" frames to store.
@@ -500,7 +667,70 @@ impl GlobalProfiler {
 
     /// How many slow "spike" frames to store.
     pub fn set_max_slow(&mut self, max_slow: usize) {
-        self.max_slow = max_slow
+        self.max_slow = max_slow;
+    }
+
+    /// Export profile data as a `.puffin` file.
+    #[cfg(feature = "serialization")]
+    pub fn save_to_path(&self, path: &std::path::Path) -> anyhow::Result<()> {
+        let mut file = std::fs::File::create(path)?;
+        self.save_to_writer(&mut file)
+    }
+
+    /// Export profile data as a `.puffin` file.
+    #[cfg(feature = "serialization")]
+    pub fn save_to_writer(&self, write: &mut impl std::io::Write) -> anyhow::Result<()> {
+        write.write_all(b"PUF0")?;
+
+        let slowest_frames = self.slowest_frames.iter().map(|f| &f.0);
+        let mut frames: Vec<_> = slowest_frames.chain(self.recent_frames.iter()).collect();
+        frames.sort_by_key(|frame| frame.frame_index);
+        frames.dedup_by_key(|frame| frame.frame_index);
+
+        for frame in frames {
+            frame.write_into(write)?;
+        }
+        Ok(())
+    }
+
+    /// Import profile data from a `.puffin` file.
+    #[cfg(feature = "serialization")]
+    pub fn load_path(path: &std::path::Path) -> anyhow::Result<Self> {
+        let mut file = std::fs::File::open(path)?;
+        Self::load_reader(&mut file)
+    }
+
+    /// Import profile data from a `.puffin` file.
+    #[cfg(feature = "serialization")]
+    pub fn load_reader(read: &mut impl std::io::Read) -> anyhow::Result<Self> {
+        let mut magic = [0_u8; 4];
+        read.read_exact(&mut magic)?;
+        if &magic != b"PUF0" {
+            anyhow::bail!("Expected .puffin magic header of 'PUF0', found {:?}", magic);
+        }
+
+        let mut slf = Self {
+            max_recent: usize::MAX,
+            ..Default::default()
+        };
+        while let Some(frame) = FrameData::read_next(read)? {
+            slf.add_frame(frame.into());
+        }
+
+        Ok(slf)
+    }
+
+    /// Call this function with each new finished frame.
+    /// The returned [`FrameSinkId`] can be used to remove the sink with [`Self::remove_sink`].
+    pub fn add_sink(&mut self, sink: FrameSink) -> FrameSinkId {
+        let id = self.next_sink_id;
+        self.next_sink_id.0 += 1;
+        self.sinks.insert(id, sink);
+        id
+    }
+
+    pub fn remove_sink(&mut self, id: FrameSinkId) -> Option<FrameSink> {
+        self.sinks.remove(&id)
     }
 }
 
@@ -547,7 +777,7 @@ impl ProfilerScope {
 
 impl Drop for ProfilerScope {
     fn drop(&mut self) {
-        ThreadProfiler::call(|tp| tp.end_scope(self.start_stream_offset))
+        ThreadProfiler::call(|tp| tp.end_scope(self.start_stream_offset));
     }
 }
 
